/**
  *  \file client/dlg-list.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-list.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/widgets.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/msgbox.h"
#include "cpluslib/holder.h"
#include "client/dlg-visual.h"

/************************** WObjectSelectionList *************************/

// FIXME: interaction with list.current() etc.

WObjectSelectionList::WObjectSelectionList(int id, int lines, int width, GObjectList& list)
    : UIListbox(id, lines, width, list.getNumObjects()), list(list)
{
    // In case first object is not selectable, bring cursor onto selectable object
    scrollTo(0, 1);
}

WObjectSelectionList::~WObjectSelectionList()
{ }

/** Fetch additional information for an object. Derived classes can override
    this to provide additional information for an object. For example,
    in proxy transfer, PCC will show current ship/ship transfers when
    selecting the proxy. */
string_t
WObjectSelectionList::getObjectInfo(GObjectId /*id*/)
{
    return string_t();
}

void
WObjectSelectionList::drawPart(GfxCanvas& can, int from, int to)
{
    const int h = getItemHeight();
    GfxParameters parm;
    parm.font = fonts[FONT_NORMAL];

    for (int line = from; line < to; ++line) {
        const int y = extent.y + h * line;
        const int index = getTop() + line;

        if (index >= list.getNumObjects()) {
            can.drawBar(GfxRect(extent.x, y, extent.w, h), standard_colors[COLOR_GRAY],
                        GFX_TRANSPARENT, SOLID_FILL, GFX_ALPHA_OPAQUE);
        } else {
            GObjectId id = list.getObjectId(index+1);
            UIColor::Color scheme = UIColor::tc_Static;
            if (id.isGameObject())
                scheme = getPlayerColor(id.getObject().getOwner());
            if (index == getCurrent()) {
                // FIXME: rewrite using drawEntry
                parm.color   = standard_colors[BLACK_COLOR_SET[scheme]];
                parm.bgcolor = getFocusState() != NoFocus ? standard_colors[COLOR_BLACK] : standard_colors[COLOR_GRAY];
            } else {
                parm.color   = standard_colors[GRAY_COLOR_SET[scheme]];
                parm.bgcolor = standard_colors[COLOR_GRAY];
            }
            if (id.getId()) {
                /* FIXME: selection, ... */
                string_t info = getObjectInfo(id);
                if (info.length()) {
                    outTextF(can, parm, extent.x, y, extent.w - info_space,
                             " " + id.getName());
                    outTextF(can, parm, extent.x + info_space, y, info_space,
                             info);
                } else {
                    outTextF(can, parm, extent.x, y, extent.w,
                             " " + id.getName());
                }
            } else {
                /* divider line */
                can.drawBar(GfxRect(extent.x, y, extent.w, h), standard_colors[COLOR_GRAY],
                            GFX_TRANSPARENT, SOLID_FILL, GFX_ALPHA_OPAQUE);
                parm.bgcolor = GFX_TRANSPARENT;
                outTextF(can, parm, extent.x + 20, y, extent.w - 40,
                         id.getName());
                parm.color = standard_colors[COLOR_SHIELD+2];
                drawRectangle(can, parm, GfxRect(extent.x, y, extent.w, h));
            }
        }
    }
}

bool
WObjectSelectionList::isAccessible(int index)
{
    return index >= 0
        && index < list.getNumObjects()
        && list.getObjectId(index+1).getId()
        && list.isExistant(index+1);
}

/** Select object /id/. \returns true iff ok, false if not. */
bool
WObjectSelectionList::selectObject(GObjectId id)
{
    int index = list.getIndexFor(id);
    if (index > 0) {
        scrollTo(index-1, 1);
        return true;
    }
    return false;
}

/** Process list change. Updates display according to new contents and also moves
    cursor to a specified object. */
void
WObjectSelectionList::onListChange(GObjectId objectToSelect)
{
    // Update bounds
    setSize(list.getNumObjects());

    // Go to object
    selectObject(objectToSelect);

    // Redraw everything
    drawWidget();
}

GObjectId
WObjectSelectionList::doStandardDialog(string_t title, int help_id)
{
    /* can we do something? */
    if (!list.getNumObjects())
        return GObjectId();

    /* okay, go ahead. */
    ObjectHolder holder;

    UIWindow win(title, UIWindow::BLUE_WINDOW, COLOR_GRAY);
    win.setLayout(UIVBoxLayout::instance5);
    win.add(*this);

    UIGroup bgroup(UIHBoxLayout::instance5);
    bgroup.add(holder.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    bgroup.add(holder.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    bgroup.add(holder.add(new UISpacer()));
    if (help_id)
        /* FIXME */
        bgroup.add(holder.add(new UIButton(cm_Help, _("Help"), 'h')).addCommand());
    win.add(bgroup);
    win.setFocusedWidget(*this);
    win.pack();
    ui_root->centerWidget(win);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);

    if (rv)
        return getCurrentObject();
    else
        return GObjectId();
}
