/**
  *  \file game/ship.cc
  *  \brief Ships
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/ship.h"
#include "util/console.h"
#include "game/spec.h"
#include "util/nls.h"
#include "game/global.h"
#include "game/planet.h"
#include "cpluslib/format.h"
#include "game/mission.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/ship-int.h"
#include "game/hullfunc.h"

/******************************* Utilities *******************************/

static const TShipXY shipxy_init = { 0, 0, 0, 0 };

static const TShip blank_ship = {
    0, 0,                                       // ID, Owner
    { UNKNOWNC, UNKNOWNC, UNKNOWNC },           // FCode
    UNKNOWNP,                                   // Warp
    0, 0,                                       // Waypoint-DX, DY
    UNKNOWNP, UNKNOWNP,                         // X, Y
    UNKNOWNP,                                   // Engine
    UNKNOWNP,                                   // Hull
    UNKNOWNP, UNKNOWNP,                         // Beam Type, Count
    UNKNOWNP,                                   // Bay Count
    UNKNOWNP,                                   // Torp Type
    UNKNOWNP,                                   // Ammo
    UNKNOWNP,                                   // Torp Launcher Count
    UNKNOWNP,                                   // Mission
    UNKNOWNP,                                   // PE
    UNKNOWNP,                                   // Tow
    UNKNOWNP, UNKNOWNP,                         // Damage, Crew
    UNKNOWNP,                                   // Colonists
    { 0 },                                      // Name
    { UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP }, // Ore
    UNKNOWNP,                                   // Supplies
    { { 0, 0, 0, 0 }, 0, 0, 0 },                // Unload
    { { 0, 0, 0, 0 }, 0, 0, 0 },                // Transfer
    UNKNOWNP,                                   // Intercept
    UNKNOWNP,                                   // Cash
};

static const TDbShipTrackEntry blank_ship_track = {
    UNKNOWNP, UNKNOWNP,                         // x, y
    UNKNOWNP,                                   // Speed
    UNKNOWNP,                                   // Heading
    UNKNOWNP                                    // Mass
};

static void
copyTargetToShip(TShip& ship, const TShipTarget& target)
{
    ship.owner       = target.owner;
    ship.warp        = target.warp;
    ship.x           = target.x;
    ship.y           = target.y;
    ship.waypoint_dx = 0;
    ship.waypoint_dy = 0;
    ship.hull        = target.hull;
    std::memcpy(ship.name, target.name, sizeof(target.name));
}

GShip::MapInfo::MapInfo(int id)
    : id(id), pos(), owner(0), scanned_mass(0), orbit(0), rc_flag(0), kind(NoShip)
{ }

/******************************** ShipInfo *******************************/

GShip::ShipInfo::ShipInfo(int id)
    : data(blank_ship),
      track_turn(0),
      heading(-1)
{
    data.ship_id = id;
    time[0] = time[1] = -1;
    for (int i = 0; i < NUM_SHIP_TRACK_ENTRIES; ++i)
        track[i] = blank_ship_track;
}

/********************************* GShip *********************************/

/** Constructor. Creates a blank ship object for ship /id/. */
GShip::GShip(int id)
    : map_info(id), ship_info(0), target_data(0),
      src_flags(), target_flags()
{
    transporter_in_use[0] = transporter_in_use[1] = false;
}

GShip::~GShip()
{
    delete ship_info;
    delete target_data;
}

GShip::ShipInfo&
GShip::createShip()
{
    if (!ship_info) {
        ship_info = new ShipInfo(map_info.id);
        // FIXME: give it a name?
        if (map_info.kind == NoShip)
            map_info.kind = HistoryShip;
    }
    return *ship_info;
}

const TShip&
GShip::getDisplayedShip() const
{
    return ship_info ? ship_info->data : blank_ship;
}

/** Adjust ship track info to start no later than turn.
    \post turn <= info.track_turn */
void
GShip::adjustShipTrack(int turn)
{
    ShipInfo& info = createShip();
    if (turn > info.track_turn) {
        int adjust = turn - info.track_turn;
        for (int i = NUM_SHIP_TRACK_ENTRIES-1; i >= 0; --i) {
            info.track[i] = (i >= adjust
                             ? info.track[i - adjust]
                             : blank_ship_track);
        }
        info.track_turn = turn;
    }
}

/* Initialisation */

/** Add SHIP.DAT entry. This assumes that if we see a ship through
    several .dat files, we get the same one each time.
    \param dat  [in] Parsed .dat file entry
    \param srcf [in] Source flags, i.e. bitfield of which file this record is from */
void
GShip::addShipData(const TShip& dat, GPlayerSet srcf)
{
    createShip().data = dat;
    // map_info.kind is updated in internalCheck()
    src_flags |= srcf;
}

/** Add TARGET.DAT entry. This assumes that if we see a ship through
    several .dat files, we get the same one each time. This does, however,
    handle the case that one record is seen through PHost's ship name filter
    and the other one isn't, and uses the better ship name.
    \param target [in] Parsed TARGET.DAT file entry
    \param srcf   [in] Source flags, i.e. bitfield of which file this record is from */
void
GShip::addTargetData(const TShipTarget& target, GPlayerSet srcf)
{
    // map_info.kind is updated in internalCheck(), which also performs
    // the target-to-ship copy if the ship isn't seen through another means.
    if (!target_data) {
        /* creating a new target */
        target_data = new TShipTarget(target);
    } else {
        /* if the new target is better than the one we have, use that */
        if (isDummyName(target_data->name, map_info.id) && !isDummyName(target.name, map_info.id))
            *target_data = target;
    }
    target_flags |= srcf;
}

/** Add SHIPXY.DAT entry. */
void
GShip::addShipXYData(TShipXY xy)
{
    map_info.pos.x = xy.x;
    map_info.pos.y = xy.y;
    map_info.owner = xy.owner;
    map_info.scanned_mass = xy.mass;
}

/** Add remote control flag.
    \param flag=0 for non-RC'd ships
    \param flag<0 if RC is forbidden
    \param flag>0 if to report the ship's real owner */
void
GShip::addRCEntry(int flag)
{
    map_info.rc_flag = flag;
}

/** Add ship history data. */
void
GShip::addShipHistoryData(const TDbShip& ship)
{
    if (src_flags.empty()) {
        /* Only accept history when no current data present */
        ShipInfo& info = createShip();
        info.data = ship.ship;
        info.time[0] = ship.time[0];
        info.time[1] = ship.time[1];
    }
}

/** Add ship track data.
    \param entry [in] Ship Track entry from database
    \param turn  [in] Turn number (absolute value) */
void
GShip::addShipTrackEntry(const TDbShipTrackEntry& entry, int turn)
{
    if (entry.x != UNKNOWNP && entry.y != UNKNOWNP) {
        /* Adjust ship position */
        adjustShipTrack(turn);

        /* Fill in new entry if possible. */
        ShipInfo& info = createShip();
        int slot = info.track_turn - turn;
        if (slot < NUM_SHIP_TRACK_ENTRIES)
            info.track[slot] = entry;
    }
}

/** Add ship function.
    \param function Identifier of function. Can be a basic hull function identifier
    (hf_XXX), or a host function Id (see GHullFunctionData::getIdFromHostId). */
void
GShip::addShipFunction(int function)
{
    special_functions.push_back(function);
}

/** Get ship data.
    \param the_data [out] Ship data to store in SHIP.DAT file */
void
GShip::getShipData(TShip& the_data) const
{
    ASSERT(ship_info);
    the_data = ship_info->data;
}

/** Get header for ship track data in history file.
    \param header [out] Header
    \return Number of track entries (0=none, no header returned in this case) */
int
GShip::getShipTrackHeader(TDbShipTrack& header) const
{
    if (ship_info) {
        header.id = map_info.id;
        header.ref_turn = ship_info->track_turn;
        return NUM_SHIP_TRACK_ENTRIES;
    } else {
        return 0;
    }
}

/** Get ship track entry by slot.
    \param entry [out] Entry
    \param slot [in] Slot to access, [0, getShipTrackHeader(...)) */
void
GShip::getShipTrackEntry(TDbShipTrackEntry& entry, int slot) const
{
    ASSERT(slot >= 0 && slot < NUM_SHIP_TRACK_ENTRIES);
    ASSERT(ship_info != 0);
    entry = ship_info->track[slot];
}

/** Get ship history data.
    \param the_data [out] Data to store in history file */
void
GShip::getShipHistoryData(TDbShip& the_data) const
{
    ASSERT(ship_info);
    the_data.ship    = ship_info->data;
    the_data.time[0] = ship_info->time[0];
    the_data.time[1] = ship_info->time[1];
}

/* GMapObject interface */

string_t
GShip::getName() const
{
    // FIXME: handle unknown/blank name
    if (ship_info)
        return getBasicString(ship_info->data.name);
    else
        return "<unknown>";
}

string_t
GShip::getFullName() const
{
    if (ship_info)
        return format(_("%s (Ship #%d)"), getName(), map_info.id);
    else
        return format(_("Ship %d"), map_info.id);
}

int
GShip::getId() const
{
    return map_info.id;
}

int
GShip::getOwner() const
{
    return map_info.owner;
}

/** Check ship existance. Returns true iff the ship is visible on the
    map (i.e. exists for us at this point in time) */
bool
GShip::isExistant() const
{
    return map_info.owner != 0;
}

GPoint
GShip::getPos() const
{
    return map_info.pos;
}

/* Checking */

void
GShip::internalCheck()
{
    /* figure out what kind we are */
    if (src_flags.nonempty()) {
        /* current ship */
        ASSERT(ship_info != 0);
        map_info.kind = CurrentShip;

        /* check that our other data agrees */
        if (map_info.pos.x != ship_info->data.x || map_info.pos.y != ship_info->data.y || map_info.owner != ship_info->data.owner) {
            /* FIXME: should not warn when we're reading host data */
            console.write(LOG_WARN, format(_("Ship #%d has inconsistent XY information - correcting."), map_info.id));
            map_info.pos.x = ship_info->data.x;
            map_info.pos.y = ship_info->data.y;
            map_info.owner = ship_info->data.owner;
        }
        if (target_data != 0) {
            if (target_data->x != ship_info->data.x || target_data->y != ship_info->data.y || target_data->owner != ship_info->data.owner) {
                console.write(LOG_WARN, format(_("Ship #%d has inconsistent target information - deleting."), map_info.id));
                delete target_data;
                target_data = 0;
            }
        }
    } else if (target_flags.nonempty()) {
        /* current target */
        ASSERT(target_data != 0);
        map_info.kind = CurrentTarget;

        /* check that our other data agrees */
        if (map_info.pos.x != target_data->x || map_info.pos.y != target_data->y || map_info.owner != target_data->owner) {
            if (map_info.owner)
                /* don't warn for FF targets. */
                console.write(LOG_WARN,
                              format(_("Ship #%d has inconsistent XY information - correcting."), map_info.id));
            map_info.pos.x = target_data->x;
            map_info.pos.y = target_data->y;
            map_info.owner = target_data->owner;
        }

        /* create ship_data entry */
        copyTargetToShip(createShip().data, *target_data);
    } else if (map_info.owner != 0) {
        /* nonvisual contact */
        map_info.kind = CurrentUnknown;

        /* create ship_data entry */
        createShip();
        ship_info->data.x     = map_info.pos.x;
        ship_info->data.y     = map_info.pos.y;
        ship_info->data.owner = map_info.owner;
    } else if (ship_info != 0) {
        /* history ship */
        map_info.kind = HistoryShip;
    } else {
        /* unknown */
        map_info.kind = NoShip;
    }
}

void
GShip::combinedCheck1(GUniverse& univ)
{
    // FIXME: handle nonexistant ships gracefully

    if (isPlayed() && !hasShipData()) {
        /* we own it but can't see it? impossible! */
        delete target_data;
        target_data = 0;
        map_info.pos.x = map_info.pos.y = map_info.owner = map_info.scanned_mass = 0;
        console.write(LOG_WARN,
                      format(_("Ship #%d is ours but we don't see it -- deleted."), map_info.id));
        // FIXME: what if during combined check someone else sees that we play this ship?
        setIsPlayed(false);
        map_info.kind = ship_info ? HistoryShip : NoShip;
        return;
    }

    /* This ship looks valid. Compute orbit. */
    map_info.orbit = univ.getPlanetAt(getPos());
    if (map_info.orbit)
        univ.getPlanet(map_info.orbit).setShipInOrbit(*this);

    /* Update ages */
    if (hasShipData()) {
        ship_info->time[0] = ship_info->time[1] = univ.getTurnNumber();
    }

    /* Update ship track. The simplest way is to generate a whole new record
       and have the regular code assimilate it. */
    if (isExistant()) {
        TDbShipTrackEntry e;
        e.x       = map_info.pos.x;
        e.y       = map_info.pos.y;
        e.speed   = getWarp().getRawValue();
        e.heading = getHeading().getRawValue();
        e.mass    = getMass();
        addShipTrackEntry(e, univ.getTurnNumber());
    }

#if 0
    /* FIXME: I'm not sure whether this code is actually useful. */
    /* Update ship track headings */
    if (ship_info != 0) {
        for (int i = 1; i < NUM_SHIP_TRACK_ENTRIES; ++i) {
            /* We update the heading if two adjoining entries are reliably
               known. This means they have a mass of strictly more than 0
               (mass 0 means a guessed target). */
            TDbShipTrackEntry& neu = ship_info->track[i-1];
            TDbShipTrackEntry& old = ship_info->track[i];
            if (neu.x > 0 && neu.mass > 0 && old.x > 0 && old.mass > 0) {
                if (neu.x == old.x && neu.y == old.y) {
                    neu.heading = -1;
                } else {
                    neu.heading = roundToInt(getHeadingDeg(neu.x - old.x, neu.y - old.y));
                }
            }
        }
    }
#endif
}

/** Check whether we have full, playable data. */
bool
GShip::hasShipData() const
{
    return map_info.kind == CurrentShip;
}

/** Check whether we have any data about this ship. */
bool
GShip::hasAnyData() const
{
    return ship_info != 0;
}

/** Check whether this ship is reliably visible (to a player). A ship can
    be unreliably visible if it guessed. It can also be reliably visible to
    one player but not another one if they are not allied.

    \param for_player Player to ask question for: is this ship known
    to that player, and will host accept orders relating to it? If zero,
    check whether ship is seen reliably by anyone. */
bool
GShip::isReliablyVisible(int for_player) const
{
    GPlayerSet set = target_flags | src_flags;
    if (!for_player)
        return set.nonempty();
    else
        return set.contains(for_player);
}

/** Get kind of this ship. */
GShip::Kind
GShip::getKind() const
{
    return map_info.kind;
}

/** Get ship source flags. This is the set of players whose SHIP
    file contains a copy of this ship (usually a unit set). */
GPlayerSet
GShip::getSrcFlags() const
{
    return src_flags;
}

/** Get target source flags. This is the set of players whose TARGET
    file contains a copy of this ship, and who therefore reliably know
    about this ship's existance. Should be disjoint from getSrcFlags().

    FIXME: what about nonvisual contacts (SHIPXY)? */
GPlayerSet
GShip::getTargetFlags() const
{
    return target_flags;
}

/* Accessors */

/** Get ship mass. Returns zero if unknown. */
int
GShip::getMass() const
{
    if (map_info.kind == CurrentShip)
        return getShipMass(ship_info->data);
    else if (map_info.kind == CurrentTarget || map_info.kind == CurrentUnknown)
        return map_info.scanned_mass;
    else
        return 0;
}

/** Get hull Id. Returns zero if unknown. */
int
GShip::getHullId() const
{
    // FIXME: convert to use mp16_t
    if (ship_info && ship_info->data.hull > 0)
        return ship_info->data.hull;
    else
        return 0;
}

/** Get hull object. */
GHull&
GShip::getHull() const
{
    int n = getHullId();
    ASSERT(n > 0);
    return ::getHull(n);
}

/** Get planet we're orbiting.
    \return planet id, 0 if none. */
int
GShip::getOrbit() const
{
    return map_info.orbit;
}

/** Get real owner of ship. */
int
GShip::getRealOwner() const
{
    if (map_info.rc_flag > 0)
        return map_info.rc_flag;
    else
        return getOwner();
}

/** Get ship's remote control flag. */
int
GShip::getRemoteControlFlag() const
{
    return map_info.rc_flag;
}

/** Check whether this ship can do special function.
    \param basic_function A basic ship function, e.g. hf_Cloak. */
bool
GShip::canDoSpecial(int basic_function) const
{
    /* Check class functions */
    if (getHull().canDoSpecial(basic_function, getRealOwner(), true,
                               GExpLevelSet::allUpTo(NUM_EXPERIENCE_LEVELS)))
        return true;

    /* Check this ship's functions */
    const GHullFunctionData& fd = GHullFunctionData::instance;
    if (special_functions.size() != 0) {
        /* Ship has special functions. Figure out levels. */
        int level = unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpLevel));
        if (level < 0)
            level = 0;
        GExpLevelSet allLevels = GExpLevelSet::allUpTo(config.NumExperienceLevels());
        GExpLevelSet thisLevel = GExpLevelSet(level);

        /* Check functions */
        for (std::size_t i = 0; i < special_functions.size(); ++i) {
            GHullFunction f;
            if (fd.getFunctionDefinition(fd.getIdFromHostId(special_functions[i]), f)) {
                if (f.getLevels().contains(allLevels) || f.getLevels().containsAnyOf(thisLevel)) {
                    return true;
                }
            }
        }
    }

    /* Not found */
    return false;
}

/** Enumerate this ship's special functions (only functions assigned to this
    ship, not including class functions). */
void
GShip::enumShipSpecificSpecials(GHullFunctionList& list) const
{
    const GHullFunctionData& fd = GHullFunctionData::instance;
    for (std::size_t i = 0; i < special_functions.size(); ++i) {
        GHullFunction f;
        if (fd.getFunctionDefinition(fd.getIdFromHostId(special_functions[i]), f)) {
            list.add(f);
        }
    }
}

/** Check whether this ship has any ship-specific functions. */
bool
GShip::hasAnyShipSpecificSpecials() const
{
    return special_functions.size() != 0;
}

/* Accessors: Course */

/** Get waypoint.
    \pre hasShipData() */
GPoint
GShip::getWaypoint() const
{
    ASSERT(map_info.kind == CurrentShip);
    return GPoint(map_info.pos.x + ship_info->data.waypoint_dx,
                  map_info.pos.y + ship_info->data.waypoint_dy);
}

/** Set waypoint.
    \pre hasShipData() */
void
GShip::setWaypoint(GPoint pt)
{
    if (ship_info != 0 && pt != getWaypoint()) {
        ship_info->data.waypoint_dx = pt.x - map_info.pos.x;
        ship_info->data.waypoint_dy = pt.y - map_info.pos.y;
        markDirty();
    }
}

/** Get waypoint X displacement.
    \pre hasShipData() */
int
GShip::getWaypointDX() const
{
    ASSERT(map_info.kind == CurrentShip);
    return ship_info->data.waypoint_dx;
}

/** Get waypoint Y displacement.
    \pre hasShipData() */
int
GShip::getWaypointDY() const
{
    ASSERT(map_info.kind == CurrentShip);
    return ship_info->data.waypoint_dy;
}

/** Get ship's heading vector.
    \return value from [0,360), corresponding to VGAP's heading angles.
    Returns unknown value if ship doesn't move or heading not known */
mp16_t
GShip::getHeading() const
{
    if (map_info.kind == CurrentShip) {
        if (ship_info->data.waypoint_dx == 0 && ship_info->data.waypoint_dy == 0) {
            return mp16_t();
        } else {
            return mp16_t(int(getHeadingDeg(ship_info->data.waypoint_dx, ship_info->data.waypoint_dy)));
        }
    } else if (map_info.kind == CurrentTarget) {
        return mp16_t(target_data->heading);
    } else {
        return mp16_t();
    }
}

/** Get warp factor.
    \return warp factor or unknown */
mp16_t
GShip::getWarp() const
{
    if (map_info.kind == CurrentShip || map_info.kind == CurrentTarget)
        return mp16_t(ship_info->data.warp);
    else
        return mp16_t();
}

/** Set warp factor. */
void
GShip::setWarp(int warp)
{
    // FIXME: doesn't create history ships - should it?
    if (ship_info != 0 && warp != getWarp()) {
        ship_info->data.warp = warp;
        markDirty();
    }
}

/** Check whether ship is hyperdriving.
    \return true iff ship has hyperdrive and tries to activate it */
bool
GShip::isHyperdriving() const
{
    return map_info.kind == CurrentShip
        && canDoSpecial(hf_Hyperdrive)
        && ship_info->data.warp > 0
        && std::memcmp("HYP", ship_info->data.fcode, 3) == 0;
}

/* Accessors: equipment */

/** Get engine type.
    \return engine type [1,NUM_ENGINES], or unknown */
mp16_t
GShip::getEngine() const
{
    return mp16_t(getDisplayedShip().engine);
}

/** Set engine type (for history).
    \param engine engine type [1,NUM_ENGINES], or unknown */
void
GShip::setEngine(mp16_t engine)
{
    createShip().data.engine = engine.getRawValue();
    markDirty();
}

/** Get beam type.
    \return beam type [1,NUM_BEAMS], or unknown */
mp16_t
GShip::getBeamType() const
{
    return mp16_t(getDisplayedShip().beam_type);
}

/** Set beam type (for history).
    \param type beam type [1,NUM_BEAMS], or unknown */
void
GShip::setBeamType(mp16_t type)
{
    createShip().data.beam_type = type.getRawValue();
    markDirty();
}

/** Get beam count.
    \return beam count, or unknown */
mp16_t
GShip::getBeamCount() const
{
    return mp16_t(getDisplayedShip().beam_count);
}

/** Set beam count (for history).
    \param count beam count, or unknown */
void
GShip::setBeamCount(mp16_t count)
{
    createShip().data.beam_count = count.getRawValue();
    markDirty();
}

/** Get bay count.
    \return bay count, or unknown */
mp16_t
GShip::getBayCount() const
{
    return mp16_t(getDisplayedShip().bay_count);
}

/** Set bay count (for history).
    \param count bay count, or unknown */
void
GShip::setBayCount(mp16_t count)
{
    createShip().data.bay_count = count.getRawValue();
    markDirty();
}

/** Get torpedo type.
    \return torpedo type [1,NUM_TORPS], or unknown */
mp16_t
GShip::getTorpType() const
{
    return mp16_t(getDisplayedShip().torp_type);
}

/** Set torpedo type (for history).
    \param type beam type [1,NUM_TORPS], or unknown */
void
GShip::setTorpType(mp16_t type)
{
    createShip().data.torp_type = type.getRawValue();
    markDirty();
}

/** Get torpedo launcher count.
    \return torpedo launcher count, or unknown */
mp16_t
GShip::getTorpLauncherCount() const
{
    return mp16_t(getDisplayedShip().torp_launcher_count);
}

/** Set torpedo launcher count (for history).
    \param count torpedo launcher count, or unknown */
void
GShip::setTorpLauncherCount(mp16_t count)
{
    createShip().data.torp_launcher_count = count.getRawValue();
    markDirty();
}

/* Accessors: Mission */

/** Set ship name.
    \param str New name */
void
GShip::setName(const string_t& str)
{
    storeBasicString(createShip().data.name, str);
    markDirty();
}

/** Get ship mission.
    \return mission number, or unknown */
mp16_t
GShip::getMission() const
{
    if (hasShipData())
        return mp16_t(getDisplayedShip().mission);
    else
        return mp16_t();
}

/** Set ship mission.
    \param m Mission number
    \param i Intercept number
    \param t Tow number

    FIXME: should this take a TMissionData object? */
void
GShip::setMission(int m, int i, int t)
{
    // FIXME: doesn't create history ships - should it? Should this work at all on history ships?
    if (ship_info == 0)
        return;

    if (m != ship_info->data.mission || i != ship_info->data.mission_intercept || t != ship_info->data.mission_tow) {
        ship_info->data.mission = m;
        ship_info->data.mission_intercept = i;
        ship_info->data.mission_tow = t;
        markDirty();

        /* FIXME: I do not exactly like this, because it's slow, and
           it references getDisplayedTurn(). Can we do better? */
        if (const GMission* msn = GMissionList::getInstance().getMissionByNumber(m, GPlayerSet(raceId(getRealOwner())))) {
            if (msn->getFlags() & GMission::mf_Waypoint) {
                if (i > 0 && i <= NUM_SHIPS && getDisplayedTurn().getShip(i).isExistant()) {
                    setWaypoint(getDisplayedTurn().getShip(i).getPos());
                }
            }
        }
    }
}

/** Get tow number. */
mp16_t
GShip::getTowId() const
{
    if (hasShipData())
        return mp16_t(getDisplayedShip().mission_tow);
    else
        return mp16_t();
}

/** Get intercept number. */
mp16_t
GShip::getIntId() const
{
    if (hasShipData())
        return mp16_t(getDisplayedShip().mission_intercept);
    else
        return mp16_t();
}

/** Get primary enemy. */
mp16_t
GShip::getPE() const
{
    return mp16_t(getDisplayedShip().primary_enemy);
}

/** Set primary enemy.
    \param pe New primary enemy, [0,NUM_PLAYERS] */
void
GShip::setPE(int pe)
{
    // FIXME: doesn't create history ships - should it?
    if (ship_info != 0 && ship_info->data.primary_enemy != pe) {
        ship_info->data.primary_enemy = pe;
        markDirty();
    }
}

/** Get ship damage.
    \return damage level, or unknown */
mp16_t
GShip::getDamage() const
{
    return mp16_t(getDisplayedShip().damage);
}

/** Set ship damage.
    \param damage damage level, or unknown */
void
GShip::setDamage(mp16_t damage)
{
    createShip().data.damage = damage.getRawValue();
    markDirty();
}

/** Get ship crew.
    \return number of crewmen, or unknown */
mp16_t
GShip::getCrew() const
{
    return mp16_t(getDisplayedShip().crew);
}

/** Set ship crew.
    \param crew number of crewmen, or unknown */
void
GShip::setCrew(mp16_t crew)
{
    createShip().data.crew = crew.getRawValue();
    markDirty();
}

/** Get friendly code. */
string_t
GShip::getFCode() const
{
    return getBasicString(getDisplayedShip().fcode);
}

/** Set friendly code. */
void
GShip::setFCode(const string_t& str)
{
    storeBasicString(createShip().data.fcode, str);
    markDirty();
}

/* Accessors: cargo */

/* FIXME: how to handle cargo of history ships? We return int for the
   time being, as that's what planets do. */

int
GShip::getAmmo() const
{
    // FIXME: handle overload etc?
    GCargoType type = getBayCount() > 0 ? el_Fighters : GCargoType(el_Torps + getTorpType() - 1);
    return ship_info->data.ammo - cargo.getReservedCargo(type);
}

int
GShip::getColonists() const
{
    return getCargo(el_Colonists);
}

int
GShip::getSupplies() const
{
    return getCargo(el_Supplies);
}

int
GShip::getMoney() const
{
    return getCargo(el_Money);
}

/** Get cargo amount sans reservations. */
int
GShip::getCargo(GCargoType type) const
{
    return getCargoRaw(type) - cargo.getReservedCargo(type);
}

/** Get cargo which is physically aboard the ship. */
int
GShip::getCargoRaw(GCargoType type) const
{
    if (!ship_info)
        return 0;

    switch (type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        return ship_info->data.ore_cargo[type];
     case el_Fighters:
        if (ship_info->data.bay_count > 0)
            return ship_info->data.ammo;
        return 0;
     case el_Colonists:
        return ship_info->data.colonists;
     case el_Supplies:
        return ship_info->data.supplies;
     case el_Money:
        return ship_info->data.money;
     default:
        ASSERT(type >= el_Torps && type < el_COUNT);
        if (ship_info->data.torp_type > 0 && getTorpType() == (type - el_Torps + 1))
            return ship_info->data.ammo;
        return 0;
    }
}

/** Change amount that is on ship. */
void
GShip::changeCargoRaw(GCargoType type, int amount)
{
    if (!amount)
        return;

    // FIXME: we need a way to directly set a ship's cargo, for
    // foreign ships. This is needed for various scans:
    // - web drain complete (=set N to 0)
    // - combat result (=set ammo)
    ASSERT(hasShipData());
    markDirty();
    switch (type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        ship_info->data.ore_cargo[type] += amount;
        break;
     case el_Fighters:
        ship_info->data.ammo += amount;
        break;
     case el_Colonists:
        ship_info->data.colonists += amount;
        break;
     case el_Supplies:
        ship_info->data.supplies += amount;
        break;
     case el_Money:
        ship_info->data.money += amount;
        break;
     default:
        ASSERT(type >= el_Torps && type < el_COUNT);
        ship_info->data.ammo += amount;
        break;
    }
}

int
GShip::getFreeCargo() const
{
    if (map_info.kind == CurrentShip)
        return getHull().getCargo()
            - getCargo(el_Tritanium)
            - getCargo(el_Duranium)
            - getCargo(el_Molybdenum)
            - getAmmo()
            - getColonists()
            - getSupplies();
    else
        return 0;
}

/* Accessors: Transporters */

/** Access a transporter.
    \param which which_one transporter to access. */
TShipTransfer&
GShip::getTransporter(Transporter which_one) throw()
{
    return which_one == t_Transfer ? ship_info->data.transfer : ship_info->data.unload;
}

/** Access a transporter.
    \param which which_one transporter to access. */
const TShipTransfer&
GShip::getTransporter(Transporter which_one) const throw()
{
    return which_one == t_Transfer ? ship_info->data.transfer : ship_info->data.unload;
}

/** Check whether a transporter is active. An active transporter has a
    full transfer order. */
bool
GShip::isTransporterActive(Transporter which_one) const throw()
{
    if (hasShipData()) {
        const TShipTransfer& t = getTransporter(which_one);
        return (t.ore[0] | t.ore[1] | t.ore[2] | t.ore[3] | t.colonists | t.supplies) != 0;
    } else {
        return false;
    }
}

/** Check whether a transporter is cancellable. A transporter is
    cancellable if its content can be moved back into main cargo room
    without overloading the ship. */
bool
GShip::isTransporterCancellable(Transporter which_one) const throw()
{
    // FIXME: this does not guarantee correctness in all cases: Take a
    // ship with 10 cargo room left, and 10 kt in the transporter. A
    // pending transaction adds 10 to the ship (i.e. it got the
    // guarantee that it'll be able to add this). When we now do
    // cancelTranporter() and commit that transaction, the ship is
    // overloaded.
    if (hasShipData()) {
        const TShipTransfer& t = getTransporter(which_one);
        long sum = (long) t.ore[0] + t.ore[1] + t.ore[2] + t.ore[3] + t.supplies + t.colonists;
        return sum == 0 || sum <= getFreeCargo();
    } else {
        return false;
    }
}

/** Cancel a transporter. Moves its contents back to main cargo room.
    \pre isTransporterCancellable(which_one) */
void
GShip::cancelTranporter(Transporter which_one) throw()
{
    ASSERT(isTransporterCancellable(which_one));
    TShipTransfer& t = getTransporter(which_one);
    ship_info->data.ore_cargo[0] += t.ore[0];
    ship_info->data.ore_cargo[1] += t.ore[1];
    ship_info->data.ore_cargo[2] += t.ore[2];
    ship_info->data.ore_cargo[3] += t.ore[3];
    ship_info->data.colonists += t.colonists;
    ship_info->data.supplies += t.supplies;
    t.ore[0] = t.ore[1] = t.ore[2] = t.ore[3] = t.supplies = t.colonists = t.id = 0;

    markDirty();
}

/** Check whether transporter is in use. A transporter is in use if
    the user is currently preparing a transport order. This means we
    must deny all other attempts of starting to prepare such an
    order. */
bool
GShip::isTransporterInUse(Transporter which_one) const throw()
{
    return transporter_in_use[which_one];
}

/** Set transporter in-use mark.
    \see isTransporterInUse */
void
GShip::setTransporterInUse(Transporter which_one, bool in_use) throw()
{
    transporter_in_use[which_one] = in_use;
}

/** Check whether ship is allowed to change waypoint. */
bool
GShip::getWaypointPermission() const
{
    // FIXME: real fleet implementation here
    return true;
}

/** Check whether ship wants to be cloned. This does not whether
    cloning actually is permitted (registeredness, hull restrictions,
    player restrictions). */
bool
GShip::isCloningAt(const GPlanet& planet) const
{
    return ship_info
        && planet.getPos() == getPos()
        && (ship_info->data.fcode[0] == 'c' && ship_info->data.fcode[1] == 'l' && ship_info->data.fcode[2] == 'n');
}

void
GShip::leaveFleet()
{
    // FIXME: real implementation here
}

/** Get ship history entry for a turn.
    \param turn Turn number */
const TDbShipTrackEntry&
GShip::getHistoryEntry(int turn) const
{
    if (ship_info != 0 && turn <= ship_info->track_turn && turn > ship_info->track_turn - NUM_SHIP_TRACK_ENTRIES) {
        return ship_info->track[ship_info->track_turn - turn];
    } else {
        return blank_ship_track;
    }
}

/******************************* GShipType *******************************/

GShipType::GShipType()
    : GIndexedObjectType(NUM_SHIPS)
{ }

bool
GShipType::isExistant(int index)
{
    GUniverse* u = getUniverse();
    if (u != 0 && index > 0 && index <= NUM_SHIPS) {
        GShip& s = u->getShip(index);
        if (s.isExistant() && s.isPlayed())
            return true;
    }
    return false;
}

GShip&
GShipType::getObject(int index)
{
    return getUniverse()->getShip(index);
}

/***************************** GAnyShipType ******************************/

GAnyShipType::GAnyShipType()
    : GIndexedObjectType(NUM_SHIPS)
{ }

bool
GAnyShipType::isExistant(int index)
{
    GUniverse* u = getUniverse();
    if (u != 0 && index > 0 && index <= NUM_SHIPS) {
        GShip& s = u->getShip(index);
        if (s.isExistant())
            return true;
    }
    return false;
}

GShip&
GAnyShipType::getObject(int index)
{
    return getUniverse()->getShip(index);
}
