/**
  *  \file int/value.cc
  *  \brief Interpreter: Values
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <math.h>
#include "int/value.h"
#include "cpluslib/format.h"
#include "io/streams.h"
#include "io/io.h"
#include "util/except.h"
#include "util/nls.h"
#include "defs.h"

const TStructureParser IntTagNode::parser[] = {
    { T_INT16, 1, offsetof(IntTagNode, tag), 0, 0 },
    { T_INT32, 1, offsetof(IntTagNode, value), 0, 0 },
    { T_END, 0, 0, 0, 0 }
};

/******************************** IntValue *******************************/

IntValue::~IntValue()
{ }

IntValue*
IntValue::load(const IntTagNode& sv, Stream& aux)
{
    switch (sv.tag) {
     case IntTagNode::Tag_Empty:
        return 0;

     case IntTagNode::Tag_Integer:
     case IntTagNode::Tag_Boolean:
        return new IntIntValue(sv, aux);

     case IntTagNode::Tag_String:
     case IntTagNode::Tag_LongString:
        return new IntStringValue(sv, aux);

     case IntTagNode::Tag_32bitFP:
     case IntTagNode::Tag_FPZero:
        return new IntFloatValue(sv, aux);

     default:
        if ((sv.tag & 0x00FF) == 0)
            throw FileFormatException(aux, _("Invalid value in data segment; file probably written by newer version of PCC"));
        return new IntFloatValue(sv, aux);
    }
}


/****************************** IntIntValue ******************************/

IntIntValue::IntIntValue(int32 value, bool is_bool)
    : value(value),
      is_bool(is_bool)
{ }

IntIntValue::IntIntValue(const IntTagNode& sv, Stream& /*aux*/)
    : value(sv.value),
      is_bool(sv.tag == sv.Tag_Boolean)
{ }

IntIntValue::~IntIntValue()
{ }

string_t
IntIntValue::toString(bool readable)
{
    if (is_bool) {
        if (readable)
            return value ? "True" : "False";
        else
            return value ? "YES" : "NO";
    } else {
        return format("%d", value);
    }
}

void
IntIntValue::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = is_bool ? sv.Tag_Boolean : sv.Tag_Integer;
    sv.value = value;
}

IntIntValue*
IntIntValue::clone()
{
    return new IntIntValue(*this);
}

/***************************** IntStringValue ****************************/

IntStringValue::IntStringValue(string_t value)
    : value(value)
{ }

IntStringValue::IntStringValue(const IntTagNode& sv, Stream& aux)
    : value()
{
    if (sv.tag == sv.Tag_String) {
        // Short String
        if (sv.value != 0) {
            if (!getPascalString(aux, value, false))
                throw FileTooShortException(aux);
        }
    } else {
        // Long String
        uint32 total = sv.value;
        while (total > 0) {
            char tmp[1024];
            uint32 now = total > sizeof(tmp) ? sizeof(tmp) : total;
            aux.readT(tmp, now);
            value.append(tmp, now);
            total -= now;
        }
    }
}

IntStringValue::~IntStringValue()
{ }

string_t
IntStringValue::toString(bool readable)
{
    if (readable) {
        if (value.find_first_of("\"\\") == value.npos) {
            // No meta-characters, use unquoted double-quote string
            return "\"" + value + "\"";
        } else if (value.find('\'') == value.npos) {
            // Double-quotes or backslashes, but no apostrophes
            return "'" + value + "'";
        } else {
            // Sufficiently complicated, so add quotes.
            string_t output = "\"";
            string_t::size_type i = 0, j;
            while (((j = value.find_first_of("\"\\", i)) != value.npos)) {
                output.append(value, i, j-i);
                output.append(1, '\\');
                output.append(value, j, 1);
                i = j+1;
            }
            output.append(value, i, value.npos);
            output += "\"";
            return output;
        }
    } else {
        return value;
    }
}

void
IntStringValue::store(IntTagNode& sv, Stream& aux)
{
    if (value.size() == 0) {
        // Empty string. Use Short String format, which is supported by PCC 1.0.8+.
        sv.tag   = sv.Tag_String;
        sv.value = 0;
    } else if (value.size() <= 255) {
        // Use Short String format, which is supported by PCC 1.0.8+.
        sv.tag   = sv.Tag_String;
        sv.value = 0x11111111;
        storePascalString(aux, value, false); // FIXME: recoding?
    } else {
        // Use Long String format, PCC 1.0.18+.
        sv.tag   = sv.Tag_LongString;
        sv.value = value.size();
        aux.write(value.data(), value.size()); // FIXME: recoding?
    }
}

IntStringValue*
IntStringValue::clone()
{
    return new IntStringValue(*this);
}

/***************************** IntFloatValue *****************************/

/** Convert REAL (real48) into double.
    \param sv contains a REAL */
static double
getFP48(const IntTagNode& sv)
{
    // REAL format:                                      FLOAT format:
    //   8 bit exponent "e"                               23 bit mantissa "m"
    //  39 bit mantissa "m"                                8 bit exponent "e"
    //   1 bit sign "s"                                    1 bit sign "s"
    // Value is e=0 => 0.0                               e=0   => (-1)^s * (2^-126) * (0.m)
    //          e>0 => (-1)^s * 2^(e-129) * (1.m)        e=255 => Inf, NaN
    //                                                   sonst => (-1)^s * 2^(e-127) * (1.m)

    int e = sv.tag & 255;
    if (e == 0) {
        return 0.0;
    } else {
        int s    = (sv.value & 0x80000000U) ? -1 : +1;
        /* Generate the mantissa 1.m times 2**39 */
        double m = (sv.value & 0x7FFFFFFF) * 256.0 + (sv.tag >> 8) + 549755813888.0 /* 2**39 */;
        return s * ldexp(m, e - (129 + 39));
    }

#if 0
    // The following converts the REAL into a FLOAT by manipulating the bit
    // representation. It is less precise, as it generates only 23 mantissa
    // bits (i.e. loses 16 bits).
    uint32 e = sv.tag & 255;           // REAL exponent
    uint32 m = sv.value & 0x7FFFFFFF;  // REAL mantissa, 31 bits (ignoring 8 LSBs)
    uint32 s = sv.value >> 31;         // REAL sign

    // Convert to FLOAT mantissa. We have 31 bits, want 23, so shift by 8.
    m >>= 8;

    if (e <= 2) {
        // Result will be FLOAT zero
        return 0.0;
    } else {
        // Assemble new value
        uint32 nv = m | ((e - 2) << 23) | (s << 31);
        float fv;
        memcpy(&fv, &nv, sizeof(nv));
        return fv;
    }
#endif
}

/** Store value in REAL (real48) format. REAL cannot store infinities, and has a smaller
    range than usual doubles. Those are stored as REAL-max. Underflows turn into REAL
    zeroes. REAL cannot store NaNs; those are converted into Tag_Empty. Although this
    does change the value in an observable way, it's probably the best we can do. The
    interpreter should try to avoid such values.
    \param value [in] Value
    \param sv [out] Stored value */
static void
storeFP48(double value, IntTagNode& sv)
{
    // REAL format:
    //   8 bit exponent "e"
    //  39 bit mantissa "m"
    //   1 bit sign "s"
    // Value is e=0 => 0.0
    //          e>0 => (-1)^s * 2^(e-129) * (1.m)
    // Maximum REAL is 1.111111111111111111111111111111111111111 * 2^126
    //            i.e. 170141183460314489226776631181521715200
    //            i.e. 1.7e+38
    // We cannot store infinities and NaNs.
    // - convert infinities and overflows to max REAL
    // - convert underflows to 0.0
    // - convert NaNs to EMPTY (!)
#ifdef HAVE_ISNAN
    if (isnan(value)) {
        sv.tag   = IntTagNode::Tag_Empty;
        sv.value = 0;
        return;
    }
#endif

    // Find out sign
    uint32 sign;
    if (value < 0) {
        sign  = 0x80000000U;
        value = -value;
    } else {
        sign = 0;
    }

    // First check for infinities or large values
    if (value > 1.0E+39) {
        sv.tag   = 0xFFFF;
        sv.value = 0x7FFFFFFF | sign;
    } else {
        // Reasonable value. Split into exponent and mantissa.
        int exp;
        double mant = frexp(value, &exp);
        exp -= 1;               // mant will be 0.1xxxxxxxxxxxxxx, we want 1.xxxxxxxxxx
        exp += 129;             // adjust for storage
        if (exp <= 0) {
            // Zero or underflow
            sv.tag   = 0;
            sv.value = 0;
        } else if (exp > 255) {
            // Overflow
            sv.tag   = 0xFFFF;
            sv.value = 0x7FFFFFFF | sign;
        } else {
            // It's a regular value. Extract mantissa bits.
            mant -= 0.5;           // 0.0yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyxxxxxxxx
            mant *= 4294967296.0;  // yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy.xxxxxxxx
            int32 bits1 = int32(mant);
            mant -= bits1;         // 0.xxxxxxxx
            mant *= 256.0;
            int32 bits2 = int32(mant);
            // Build result
            sv.tag   = exp | (256*bits2);
            sv.value = bits1 | sign;
        }
    }
}


IntFloatValue::IntFloatValue(double value)
    : value(value)
{ }

IntFloatValue::IntFloatValue(const IntTagNode& sv, Stream& /*aux*/)
{
    using std::memcpy;
    if (sv.tag == sv.Tag_FPZero) {
        // Zero
        value = 0.0;
    } else if (sv.tag == sv.Tag_32bitFP) {
        // 32-bit FP
        // FIXME: portability. We now assume that floats have the same endianness as int32s.
        float tmp;
        CTASSERT(sizeof(tmp) == sizeof(sv.value));
        memcpy(&tmp, &sv.value, sizeof(sv.value));
        value = tmp;
    } else {
        // 48-bit FP
        value = getFP48(sv);
    }
}

IntFloatValue::~IntFloatValue()
{ }

string_t
IntFloatValue::toString(bool /*readable*/)
{
    string_t result = format("%f", value);
    if (result.find('.') != result.npos && result.find_first_of("Ee") == result.npos) {
        string_t::size_type cut = result.size();
        while (cut > 0 && result[cut-1] == '0')
            --cut;
        if (cut > 0 && result[cut-1] == '.')
            --cut;
        result.erase(cut);
    }
    return result;
}

void
IntFloatValue::store(IntTagNode& sv, Stream& /*aux*/)
{
    using std::memcpy;
    if (value == 0.0) {
        sv.tag   = sv.Tag_FPZero;
        sv.value = 0;
    } else {
        storeFP48(value, sv);
    }
}

IntFloatValue*
IntFloatValue::clone()
{
    return new IntFloatValue(*this);
}
