/**
  *  \file ui/widget-num.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "util/misc.h"
#include "gfx/palette.h"
#include "ui/layout.h"

UINumberSelector::UINumberSelector(int id, long low, long high, int step)
    : UIWidget(id), current(low), low(low), high(high), step(step)
{ }

UINumberSelector::~UINumberSelector()
{ }

void
UINumberSelector::setValue(long val)
{
    if (val < low)
        val = low;
    if (val > high)
        val = high;
    if (val != current) {
        current = val;
        drawWidget(false);
        sig_changed.raise(getId());
    }
}

void
UINumberSelector::setRange(int min, int max)
{
    this->low = min;
    this->high = max;
    setValue(current);     /* this will check bounds and update */
}

void
UINumberSelector::increment(int n)
{
    if (n == 0)
        n = 1;
    setSelectedWidget();
    setValue(getValue() + n);
}

void
UINumberSelector::decrement(int n)
{
    if (n == 0)
        n = 1;
    setSelectedWidget();
    setValue(getValue() - n);
}

bool
UINumberSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && hasState(st_Focused)) {
        switch (e.key.code) {
         case '+':
         case SDLK_RIGHT:
            if (int n = e.key.arg)
                increment(n);
            else
                increment(step);
            return true;
         case '-':
         case SDLK_LEFT:
            if (int n = e.key.arg)
                decrement(n);
            else
                decrement(step);
            return true;
         case ss_Ctrl + '+':
         case ss_Ctrl + SDLK_RIGHT:
            increment(100);
            return true;
         case ss_Ctrl + '-':
         case ss_Ctrl + SDLK_LEFT:
            decrement(100);
            return true;
         case ss_Shift + '+':
         case ss_Shift + SDLK_RIGHT:
            increment(1);
            return true;
         case ss_Shift + '-':
         case ss_Shift + SDLK_LEFT:
            decrement(1);
            return true;
         case ss_Alt + '+':
         case ss_Alt + SDLK_RIGHT:
            setSelectedWidget();
            setValue(high);
            return true;
         case ss_Alt + '-':
         case ss_Alt + SDLK_LEFT:
            setSelectedWidget();
            setValue(low);
            return true;
        }
    } else if (e.type == ev_Mouse && e.mouse.b && !hasState(st_Focused) && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        setSelectedWidget();
        setFocusedWidget();
        return true;
    }
    return UIWidget::handleEvent(e, second_pass);
}

bool
UINumberSelector::doStandardDialog(string_t title, string_t prompt, int /*help_id*/)
{
    UIWindow      win(title, UIWindow::BLUE_WINDOW, COLOR_GRAY);
    UIStaticText  pr(prompt, UIStaticText::BlackTitle);
    UIButton      plus("+", '+'), minus("-", '-');
    UIButton      ok(1000, _("OK"), SDLK_RETURN);
    UIButton      cancel(1001, _("Cancel"), SDLK_ESCAPE);
    UIGroup       g1(UIHBoxLayout::instance5), g2(UIHBoxLayout::instance5);
    UISpacer      spacer;

    ok.addStop();
    cancel.addStop();

    g1.add(minus);
    g1.add(*this);
    g1.add(plus);

    g2.add(ok);
    g2.add(cancel);
    g2.add(spacer);

    win.setLayout(UIVBoxLayout::instance5);
    win.add(pr);
    win.add(g1);
    win.add(g2);
    win.pack();

    setFocusedWidget();

    ui_root->centerWidget(win);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);

    return rv == 1000;
}

/*************************** UIDecimalSelector ***************************/

/** \class UIDecimalSelector
    \brief A "move-or-type" number selector.

    This widget provides behaviour close to a normal input line (i.e.,
    digits "echo" and backspace deletes), but also accepts '+' or '-'
    resp. cursor keys like all number selectors.

    Internal consistency model: this maintains a consistent value all
    the time. Special states affect display only. For example, assume
    /low/=3, and value is 5. If the user deletes that using backspace,
    the internal stored value will still be 3 but the widget displays
    an empty line. If the user enters a '9' now, the value will be 9,
    not 39. There still is the problem that the user cannot enter a
    value such as 10, because 1 would be below our minimum. */

UIDecimalSelector::UIDecimalSelector(int id, long low, long high, int step)
    : UINumberSelector(id, low, high, step),
      state(sNormal)
{ }

UIDecimalSelector::~UIDecimalSelector()
{ }

bool
UIDecimalSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (hasState(st_Focused) && e.type == ev_Key) {
        if (e.key.code >= '0' && e.key.code <= '9') {
            // Digit: type
            setSelectedWidget();
            long n = e.key.code - '0';
            if (state == sNormal)
                n += 10*getValue();
            if (getMin() <= n && n <= getMax()) {
                setValue(n);
                wState(sNormal);
            }
            return true;
        } else if (e.key.code == SDLK_BACKSPACE) {
            // Backspace: delete last digit
            setSelectedWidget();
            if (getValue() < 10)
                wState(sZeroed);
            else
                wState(sNormal);
            setValue(getValue() / 10);
            return true;
        } else if (e.key.code == SDLK_DELETE) {
            // Del: clear if selected
            setSelectedWidget();
            if (state == sTypeErase) {
                setValue(0);
                wState(sZeroed);
            }
            return true;
        } else if (e.key.code == ss_Ctrl + SDLK_DELETE || e.key.code == ss_Ctrl + 'y') {
            // Ctrl-Del, Ctrl-Y: clear
            setSelectedWidget();
            setValue(0);
            wState(sZeroed);
            return true;
        } else if (e.key.code == ss_Ctrl + 'u') {
            // Ctrl-U: unselect
            setSelectedWidget();
            wState(sNormal);
            return true;
        } else if (e.key.code == ss_Ctrl + 't') {
            // Ctrl-T: Swap last two digits
            // Because we cannot have leading zeroes, this will also turn 2 into 20 and vice versa.
            setSelectedWidget();
            wState(sNormal);
            long n = getValue();
            setValue(100*(n/100)
                     + 10*(n%10)
                     + (n/10)%10);
            return true;
        }
    }

    // FIXME: We must track whether the base class' logic changes our value,
    // and adapt our state accordingly. This is a hack, but it's easier than
    // the other way (hooking into sig_changed and ensure that we never trigger
    // ourselves). Plus, PCC 1.x does the same :-)
    long l = getValue();
    bool b = UINumberSelector::handleEvent(e, second_pass);
    if (state != sNormal && l != getValue())
        wState(sNormal);
    return b;
}

void
UIDecimalSelector::onStateChange(int state, bool enabled)
{
    if (state & st_Focused) {
        if (enabled)
            wState(sTypeErase);
        else
            wState(sNormal);
        drawWidget(false);
    }
    UINumberSelector::onStateChange(state, enabled);
}

void
UIDecimalSelector::drawContent(GfxCanvas& can)
{
    UISkin& s = getSkin();
    s.startDraw();

    string_t post, val;

    if (getFocusState() != NoFocus) {
        post = "_";              // FIXME: replace by proper cursor
        s.setColor(UIColor::tc_Input);
    } else {
        s.setColor(UIColor::tc_Static);
    }

    if (hasState(pds_ShowMax))
        post += format(_(" (max. %d)"), getMax());

    if (state != sZeroed)
        val = getStringValue();

    int val_width  = fonts[FONT_TITLE]->getTextWidth(val);
    int post_width = fonts[FONT_TITLE]->getTextWidth(post);

    if (val_width >= getExtent().w)
        val_width = getExtent().w;
    if (post_width + val_width >= getExtent().w)
        post_width = getExtent().w - val_width;

    if (hasState(pds_RightJust)) {
        /* right-justified */
        const int rx = getExtent().x + getExtent().w;
        s.setSolid();
        s.setColor(UIColor::tc_Static);
        s.outTextF(can, FONT_TITLE, rx - post_width, getExtent().y, post_width, post);
        s.drawBackground(can, GfxRect(getExtent().x, getExtent().y,
                                      getExtent().w - post_width - val_width, getExtent().h));

        if (state == sTypeErase) {
            s.setSolid(false);
            drawSolidBar(can, GfxRect(rx - post_width - val_width, getExtent().y,
                                      val_width, getExtent().h),
                         s.getParameters().color);
            s.setColor(UIColor::tc_Background);
        } else if (getFocusState() != NoFocus) {
            s.setColor(UIColor::tc_Input);
        } else {
            s.setColor(UIColor::tc_Static);
        }
        s.outTextF(can, FONT_TITLE, rx - post_width - val_width, getExtent().y, val_width, val);
    } else {
        /* left-justified */
        if (state == sTypeErase) {
            s.setSolid(false);
            drawSolidBar(can, GfxRect(getExtent().x, getExtent().y,
                                      val_width, font_heights[FONT_TITLE]),
                         s.getParameters().color);
            s.setColor(UIColor::tc_Background);
        } else {
            s.setSolid();
        }

        s.outTextF(can, FONT_TITLE, getExtent().x, getExtent().y, val_width, val);
        s.setSolid();
        s.setColor(UIColor::tc_Static);
        s.outTextF(can, FONT_TITLE, getExtent().x + val_width, getExtent().y, getExtent().w - val_width, post);
    }
    can.update(getExtent());
}

void
UIDecimalSelector::wState(State st)
{
    if (st != state) {
        state = st;
        drawWidget(false);
    }
}

/** Get value as string. This should format the value of this widget
    (getValue()) as a string. Descendants can override this method to
    provide additional info. \default returns just the number */
string_t
UIDecimalSelector::getStringValue()
{
    return numToString(getValue());
}

void
UIDecimalSelector::getLayoutInfo(LayoutInfo& li)
{
    li.flags = li_FlexibleH;

    // FIXME? heuristic based on font_ems
    int width = (hasState(pds_ShowMax) ? font_ems[FONT_TITLE]*10 : font_ems[FONT_TITLE]*5);
    li.pref_size = li.min_size = GfxPoint(width, font_heights[FONT_TITLE]);
}
